import unittest

from src.otyg_risk_base.quantitative_risk import QuantitativeRisk



class TestQuantitativeRisk(unittest.TestCase):
    def test_no_arg(self):
        risk = QuantitativeRisk()
        self.assertIsInstance(risk, QuantitativeRisk)
        for s in ["threat_event_frequency","vulnerability","loss_event_frequency","loss_magnitude","annual_loss_expectancy","budget","currency"]:
            self.assertIn(s, risk.to_dict())

    def test_equality(self):
        risk = QuantitativeRisk()
        self.assertTrue(risk == risk)
    
    def test_serialization_deserialization(self):
        risk = QuantitativeRisk()
        new_risk = QuantitativeRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == new_risk)
        risk = QuantitativeRisk({'threat_event_frequency': {'min':0,'probable':1,'max':2}, 'vulnerability': {'min':1,'probable':2,'max':3}, 'loss_magnitude':{'min':3,'probable':4,'max':5}, 'budget': 10000, 'currency':"SEK"})
        new_risk_2 = QuantitativeRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == new_risk_2)
        self.assertFalse(risk == new_risk)

if __name__ == '__main__':
    unittest.main()


#
# BSD 3-Clause License
#
# Copyright (c) 2026, Martin Vesterlund
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from .qualitative_scale import QualitativeScale
from .montecarlo import MonteCarloRange
from .qualitative_risk import QualitativeRisk
from .utils import *
from .quantitative_risk import QuantitativeRisk

class HybridRisk():
    def __init__(self, values:dict = None):
        # TODO: Markörer för vad som finns i dicten
        if values:
            if "mappings" in values:
                qs = QualitativeScale(scales=values.get('mappings'))
            else:
                qs = QualitativeScale()
            self.quantitative = QuantitativeRisk(values=values)
            self.qualitative = QualitativeRisk(
            likelihood_init=qs.get(raw=self.quantitative.threat_event_frequency.probable, mapping='likelihood_initiation_or_occurence').get('numeric'),
            likelihood_impact=qs.get(raw=self.quantitative.vuln_score.probable, mapping='likelihood_adverse_impact').get('numeric'),
                impact=qs.get(raw=self.quantitative.loss_magnitude.probable, mapping='impact').get('numeric'),
                mappings=qs)
        else:
            self.quantitative = QuantitativeRisk()
            qs = QualitativeScale()
            self.qualitative = QualitativeRisk(
                likelihood_init=qs.get(raw=self.quantitative.threat_event_frequency.probable, mapping='likelihood_initiation_or_occurence').get('numeric'),
                likelihood_impact=qs.get(raw=self.quantitative.vuln_score.probable, mapping='likelihood_adverse_impact').get('numeric'),
                impact=qs.get(raw=self.quantitative.loss_magnitude.probable, mapping='impact').get('numeric'),
                mappings=qs)
    def to_dict(self):
        me = {
            "quantitative": self.quantitative.to_dict(),
            "qualitative": self.qualitative.to_dict()
        }
        return me
    @classmethod
    def from_dict(cls, values:dict):
        me = HybridRisk(values=None)
        me.qualitative = QualitativeRisk.from_dict(values=values.get('qualitative'))
        me.quantitative = QuantitativeRisk.from_dict(values=values.get('quantitative'))
        return me

    def get(self):
        return self.risk.copy()
    
    def __hash__(self):
        return hash((self.qualitative, self.quantitative))
    
    def __eq__(self, value):
        return self.__hash__() == value.__hash__()



from typing import Dict, Tuple

from .qualitative_scale import QualitativeScale
from .utils import *


class QualitativeRisk():

    def __init__(self, likelihood_init:int=1, likelihood_impact:int=1, impact:int=1, mappings:QualitativeScale=None):
        if not mappings:
            self.mappings=QualitativeScale()
        else:
            self.mappings = mappings
        raw_likelihood = self.mappings.get(raw=likelihood_impact*likelihood_init, mapping="risk")
        self.overall_likelihood:str = raw_likelihood.get("text")
        self.overall_likelihood_num:int = raw_likelihood.get("numeric")
        self.likelihood_initiation_or_occurence:str = self.get_text_from_num(likelihood_init)
        self.likelihood_adverse_impact:str = self.get_text_from_num(likelihood_impact)
        self.impact:str = self.get_text_from_num(impact)
        self.overall_risk:str = self.mappings.get(raw=self.overall_likelihood_num*impact, mapping="risk").get("text")

    def get_text_from_num(self, num):
        value = ""
        try:
            value = self.mappings.num_to_text[num]
        except KeyError:
            try:
                value = self.mappings.num_to_text[str(num)]
            except:
                raise
        return value
    def get(self):
        return {'risk': self.overall_risk, 'likelihood': self.overall_likelihood, 'impact': self.impact}
    
    def to_dict(self):
        risk = {
            "overall_likelihood": self.overall_likelihood,
            "overall_likelihood_num" : self.overall_likelihood_num,
            "likelihood_initiation_or_occurence" : self.likelihood_initiation_or_occurence,
            "likelihood_adverse_impact" : self.likelihood_adverse_impact,
            "impact" : self.impact,
            "overall_risk" : self.overall_risk,
            "mappings": self.mappings.to_dict()
        }
        return risk
    @classmethod
    def from_dict(cls, values:dict):
        risk = QualitativeRisk()
        risk.overall_likelihood = values.get("overall_likelihood")
        risk.overall_likelihood_num = values.get("overall_likelihood_num")
        risk.likelihood_initiation_or_occurence = values.get("likelihood_initiation_or_occurence")
        risk.likelihood_adverse_impact = values.get("likelihood_adverse_impact")
        risk.impact = values.get("impact")
        risk.overall_risk = values.get("overall_risk")
        risk.mappings = QualitativeScale.from_dict(scales=values.get("mappings"))
        return risk

    def __hash__(self):
        return hash(freeze(self.to_dict()))
    
    def __eq__(self, other):
        return self.__hash__() == other.__hash__()


from decimal import Decimal
import unittest

from src.otyg_risk_base.montecarlo import MonteCarloRange, MonteCarloSimulation

class TestMonteCarloRange(unittest.TestCase):
    def test_no_arg(self):
        mr = MonteCarloRange()
        self.assertEqual(mr.probable, Decimal(0))
    
    def test_probable_less_than_min(self):
        with self.assertRaises(ValueError):
            mc = MonteCarloRange(min= 1, probable=0, max=2)
    
    def test_equals(self):
        actual = MonteCarloRange(min= 0, probable=1, max=2)
        other = MonteCarloRange(min= 0, probable=1, max=2)
        other_neq = MonteCarloRange(min= 1, probable=2, max=3)
        self.assertTrue(actual == actual)
        self.assertTrue(actual == other)
        self.assertFalse(actual == other_neq)
        self.assertFalse(actual == 1)

    def test_add(self):
        actual = MonteCarloRange(min= 0, probable=1, max=2).add(MonteCarloRange(min= 1, probable=2, max=3))
        expected = MonteCarloRange(min= 1, probable=3, max=5)
        actual_scalar = MonteCarloRange(min= 0, probable=1, max=2).add(1)
        expected_scalar = MonteCarloRange(min= 1, probable=2, max=3)
        not_expected = MonteCarloRange(min= 2, probable=3, max=5)
        self.assertTrue(actual == expected)
        self.assertTrue(actual_scalar == expected_scalar)
        self.assertFalse(actual == not_expected)

    def test_serialization_and_deserialization(self):
        original = MonteCarloRange(min= 1, probable=3, max=5)
        org_dict = original.to_dict()
        copy = MonteCarloRange.from_dict(org_dict)
        self.assertEqual(original, copy)
        self.assertNotEqual(original, MonteCarloRange.from_dict({
            "min": 1,
            "probable": 3,
            "max": 10
        }))
        
class TestMonteCarloSimulation(unittest.TestCase):
    def test_creation(self):
        range = MonteCarloRange(min=1, probable=2, max=3)
        sim = MonteCarloSimulation(range=range)
        self.assertAlmostEqual(sim.min, 1, delta=0.1)
        self.assertAlmostEqual(sim.probable, 2, delta=0.1)
        self.assertAlmostEqual(sim.max, 3, delta=0.1)
        self.assertAlmostEqual(sim.p90, Decimal(2.5), delta=0.1)
    
    def test_equals(self):
        sim = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        other = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        self.assertTrue(sim == sim)
        self.assertFalse(sim == other)
    
    def test_serialization_deserialization(self):
        sim = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        other = MonteCarloSimulation().from_dict(sim.to_dict())
        self.assertTrue(sim == other)
    
    def test_multiplication(self):
        first = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        second = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        
        mul = first.multiply(second)
        mul_oper = first * second
        mul_roper = second * first
        threshold = 0.3
        self.assertAlmostEqual(mul.min, 1, delta=threshold)
        self.assertAlmostEqual(mul.min, mul_oper.min, delta=threshold)
        self.assertAlmostEqual(mul.min, mul_roper.min, delta=threshold)
        self.assertAlmostEqual(mul.probable, 4, delta=1)
        self.assertAlmostEqual(mul.probable, mul_oper.probable, delta=1)
        self.assertAlmostEqual(mul.probable, mul_roper.probable, delta=1)
        self.assertAlmostEqual(mul.max, 9, delta=threshold)
        self.assertAlmostEqual(mul.max, mul_oper.max, delta=threshold)
        self.assertAlmostEqual(mul.max, mul_roper.max, delta=threshold)
        self.assertAlmostEqual(mul.p90, Decimal(7), delta=threshold)
        self.assertAlmostEqual(mul.p90, mul_oper.p90, delta=threshold)
        self.assertAlmostEqual(mul.p90, mul_roper.p90, delta=threshold)
        scalar = 2
        mul_scalar = first.multiply(scalar)
        mul_scalar_oper = first * scalar
        mul_scalar_roper = scalar * first
        self.assertAlmostEqual(mul_scalar.min, 2, delta=threshold)
        self.assertAlmostEqual(mul_scalar.min, mul_scalar_oper.min, delta=threshold)
        self.assertAlmostEqual(mul_scalar.min, mul_scalar_roper.min, delta=threshold)
        self.assertAlmostEqual(mul_scalar.probable, 4, delta=1)
        self.assertAlmostEqual(mul_scalar.probable, mul_scalar_oper.probable, delta=1)
        self.assertAlmostEqual(mul_scalar.probable, mul_scalar_roper.probable, delta=1)
        self.assertAlmostEqual(mul_scalar.max, 6, delta=threshold)
        self.assertAlmostEqual(mul_scalar.max, mul_scalar_oper.max, delta=threshold)
        self.assertAlmostEqual(mul_scalar.max, mul_scalar_roper.max, delta=threshold)
        self.assertAlmostEqual(mul_scalar.p90, Decimal(5), delta=threshold)
        self.assertAlmostEqual(mul_scalar.p90, mul_scalar_oper.p90, delta=threshold)
        self.assertAlmostEqual(mul_scalar.p90, mul_scalar_roper.p90, delta=threshold)
    
    def test_addition(self):
        first = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        second = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        
        add = first.add(second)
        add_oper = first + second
        add_roper = second + first
        threshold = 0.3
        self.assertAlmostEqual(add.min, 2, delta=threshold)
        self.assertAlmostEqual(add.min, add_oper.min, delta=threshold)
        self.assertAlmostEqual(add.min, add_roper.min, delta=threshold)
        self.assertAlmostEqual(add.probable, 4, delta=1)
        self.assertAlmostEqual(add.probable, add_oper.probable, delta=1)
        self.assertAlmostEqual(add.probable, add_roper.probable, delta=1)
        self.assertAlmostEqual(add.max, 6, delta=threshold)
        self.assertAlmostEqual(add.max, add_oper.max, delta=threshold)
        self.assertAlmostEqual(add.max, add_roper.max, delta=threshold)
        self.assertAlmostEqual(add.p90, Decimal(5), delta=threshold)
        self.assertAlmostEqual(add.p90, add_oper.p90, delta=threshold)
        self.assertAlmostEqual(add.p90, add_roper.p90, delta=threshold)
        scalar = 2
        add_scalar = first.add(scalar)
        add_scalar_oper = first + scalar
        add_scalar_roper = scalar + first
        self.assertAlmostEqual(add_scalar.min, 3, delta=threshold)
        self.assertAlmostEqual(add_scalar.min, add_scalar_oper.min, delta=threshold)
        self.assertAlmostEqual(add_scalar.min, add_scalar_roper.min, delta=threshold)
        self.assertAlmostEqual(add_scalar.probable, 4, delta=1)
        self.assertAlmostEqual(add_scalar.probable, add_scalar_oper.probable, delta=1)
        self.assertAlmostEqual(add_scalar.probable, add_scalar_roper.probable, delta=1)
        self.assertAlmostEqual(add_scalar.max, 5, delta=threshold)
        self.assertAlmostEqual(add_scalar.max, add_scalar_oper.max, delta=threshold)
        self.assertAlmostEqual(add_scalar.max, add_scalar_roper.max, delta=threshold)
        self.assertAlmostEqual(add_scalar.p90, Decimal(4), delta=0.6)
        self.assertAlmostEqual(add_scalar.p90, add_scalar_oper.p90, delta=0.6)
        self.assertAlmostEqual(add_scalar.p90, add_scalar_roper.p90, delta=0.6)
    
    def test_delta(self):
        first = MonteCarloSimulation(range=MonteCarloRange(min=1, probable=2, max=3))
        second = MonteCarloSimulation(range=MonteCarloRange(min=2, probable=4, max=6))
        scalar = 2
        sub = first.delta(second)
        sub_scalar = first.delta(scalar)
        # 1-2 = -1, 2-4 = -2, 3-6 = -3
        threshold = 0.3
        self.assertAlmostEqual(sub.min, -3, delta=threshold)
        self.assertAlmostEqual(sub.probable, -2, delta=threshold)
        self.assertAlmostEqual(sub.max, -1, delta=threshold)
        self.assertAlmostEqual(sub_scalar.min, -1, delta=threshold)
        self.assertAlmostEqual(sub_scalar.probable, 0, delta=threshold)
        self.assertAlmostEqual(sub_scalar.max, 1, delta=threshold)

if __name__ == '__main__':
    unittest.main()


from decimal import Decimal
import unittest

from src.otyg_risk_base.montecarlo import MonteCarloRange
from src.otyg_risk_base.hybrid import HybridRisk

class TestQuantitativeRisk(unittest.TestCase):
    def test_set_qualitative(self):
        risk = HybridRisk({'threat_event_frequency': {'min':0,'probable':1,'max':2}, 'vulnerability': {'min':0,'probable':2,'max':3}, 'loss_magnitude':{'min':1,'probable':2,'max':3}, 'budget': 10000, 'currency':"SEK"})
        self.assertEqual(risk.qualitative.overall_risk, "Very High")
        self.assertEqual(risk.qualitative.overall_likelihood, "High")
        self.assertEqual(risk.qualitative.impact, "Very High")
    
    def test_equality(self):
        risk = HybridRisk({'threat_event_frequency': {'min':0,'probable':1,'max':2}, 'vulnerability': {'min':1,'probable':2,'max':3}, 'loss_magnitude':{'min':3,'probable':4,'max':5}, 'budget': 10000, 'currency':"SEK"})
        self.assertTrue(risk == risk)
        risk_mod = HybridRisk({'threat_event_frequency': {'min':1,'probable':3,'max':5}, 'vulnerability': {'min':1,'probable':2,'max':3}, 'loss_magnitude':{'min':3,'probable':4,'max':5}, 'budget': 10000, 'currency':"SEK"})
        self.assertFalse(risk == risk_mod)
    
    def test_serialization_deserialization(self):
        risk = HybridRisk({'threat_event_frequency': MonteCarloRange(min=0, probable=1, max=2), 'vulnerability':  MonteCarloRange(min=1, probable=2, max=3), 'loss_magnitude': MonteCarloRange(min=4, probable=5, max=6), 'budget': 10000, 'currency':"SEK"})
        risk_new = HybridRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == risk_new)
        risk = HybridRisk({'threat_event_frequency': {'min':0,'probable':1,'max':2}, 'vulnerability': {'min':1,'probable':2,'max':3}, 'loss_magnitude':{'min':3,'probable':4,'max':5}, 'budget': 10000, 'currency':"SEK"})
        risk_new = HybridRisk.from_dict(risk.to_dict())
        self.assertTrue(risk == risk_new)
        risk_mod = HybridRisk({'threat_event_frequency': {'min':1,'probable':3,'max':5}, 'vulnerability': {'min':1,'probable':2,'max':3}, 'loss_magnitude':{'min':3,'probable':4,'max':5}, 'budget': 10000, 'currency':"SEK"})
        risk_new = HybridRisk.from_dict(risk_mod.to_dict())
        self.assertFalse(risk == risk_new)
        self.assertTrue(risk_mod == risk_new)

    def test_ui_bugg(self):
        values = {'threat_event_frequency': {'min': 0.023, 'probable': 0.12250000000000001, 'max': 0.222},
                  'vulnerability': {'min': 0.0, 'probable': 0.085, 'max': 0.17},
                  'loss_magnitude': {'min': 0.005, 'probable': 0.01, 'max': 0.02},
                  'budget': Decimal('1000000'), 'currency': 'SEK',
                  'mappings': {
                      'likelihood_initiation_or_occurence': [{'value': 1, 'low': 0.01, 'high': 0.1}, {'value': 2, 'low': 0.1, 'high': 1}, {'value': 3, 'low': 1, 'high': 10}, {'value': 4, 'low': 10, 'high': 100}, {'value': 5, 'low': 100, 'high': 1000}],
                      'likelihood_adverse_impact': [{'value': 1, 'low': 0.01, 'high': 0.12}, {'value': 2, 'low': 0.12, 'high': 0.25}, {'value': 3, 'low': 0.25, 'high': 0.5}, {'value': 4, 'low': 0.5, 'high': 0.75}, {'value': 5, 'low': 0.75, 'high': 1}], 
                      'impact': [{'value': 1, 'low': 0.0001, 'high': 0.005}, {'value': 2, 'low': 0.005, 'high': 0.01}, {'value': 3, 'low': 0.01, 'high': 0.02}, {'value': 4, 'low': 0.02, 'high': 0.05}, {'value': 5, 'low': 0.05, 'high': 1}], 
                      'num_to_text': {'5': 'Very High', '4': 'High', '3': 'Moderate', '2': 'Low', '1': 'Very Low', '0': 'Very Low'}, 
                      'risk': [{'value': 1, 'low': 1, 'high': 5}, {'value': 2, 'low': 5, 'high': 9}, {'value': 3, 'low': 9, 'high': 13}, {'value': 4, 'low': 13, 'high': 20}, {'value': 5, 'low': 20, 'high': 26}]}}
        risk = HybridRisk(values=values)
        self.assertIsNotNone(risk.qualitative.overall_risk)



import unittest

from src.otyg_risk_base.qualitative_scale import QualitativeScale



class TestQualitativeScale(unittest.TestCase):
    def test_get(self):
        scale = QualitativeScale()
        self.assertEqual(scale.get(raw=0.0001, mapping="likelihood_initiation_or_occurence").get('numeric'), 1)
        self.assertEqual(scale.get(raw=0.1, mapping="likelihood_initiation_or_occurence").get('numeric'), 2)
        self.assertEqual(scale.get(raw=0.1, mapping="likelihood_initiation_or_occurence").get('text'), "Low")
        self.assertEqual(scale.get(raw=0.06, mapping="impact").get('text'), "Very High")
        self.assertEqual(scale.get(raw=0.74, mapping="likelihood_adverse_impact").get('text'), "High")
        self.assertEqual(scale.get(raw=16, mapping="risk").get('text'), "High")
    
    
if __name__ == '__main__':
    unittest.main()


#
# BSD 3-Clause License
#
# Copyright (c) 2026, Martin Vesterlund
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import numpy as np
from decimal import *


class MonteCarloRange():
    def __init__(self, min: Decimal = Decimal(0.00), probable: Decimal = Decimal(0.00), max: Decimal = Decimal(0.00)):
        if not (max == min) and ((probable < min) or (probable > max) or (min > max)):
            raise ValueError
        self.max = Decimal(max)
        self.min = Decimal(min)
        self.probable = Decimal(probable)
        if not self.probable.is_zero() and (self.max == self.min):
            self.max = Decimal(self.probable*2)
            self.min = Decimal(self.probable/2)
        elif self.probable.is_zero and not self.max.is_zero():
            self.probable = (self.min + self.max)/2

    def to_dict(self):
        return {
            "min": float(self.min),
            "probable": float(self.probable),
            "max": float(self.max)
        }
    
    def add(self, other = None):
        if isinstance(other, MonteCarloRange):
            max = self.max + other.max
            min = self.min + other.min
            probable = self.probable + other.probable
        else:
            max = self.max + other
            min = self.min + other
            probable = self.probable + other
        result = MonteCarloRange(min=Decimal(min), probable=Decimal(probable), max=Decimal(max))
        return result

    def sub(self, other = None):
        if isinstance(other, MonteCarloRange):
            max = self.max - other.max
            min = self.min - other.min
            probable = self.probable - other.probable
        else:
            max = self.max - other
            min = self.min - other
            probable = self.probable - other
        if max < min:
            tmp = max
            max = min
            min = tmp
        result = MonteCarloRange(min=Decimal(min), probable=Decimal(probable), max=Decimal(max))
        return result
    
    def multiply(self, other = None):
        if isinstance(other, MonteCarloRange):
            max = self.max * other.max
            min = self.min * other.min
            probable = self.probable * other.probable
        else:
            max = Decimal(self.max) * Decimal(other)
            min = Decimal(self.min) * Decimal(other)
            probable = Decimal(self.probable) * Decimal(other)
        result = MonteCarloRange(min=Decimal(min), probable=Decimal(probable), max=Decimal(max))
        return result

    @classmethod
    def from_dict(cls, values:dict=None):
        if isinstance(values, dict):
            return MonteCarloRange(min=values['min'], probable=values['probable'], max=values['max'])
        elif isinstance(values, MonteCarloRange):
            return values
        else:
            raise TypeError

    def __repr__(self):
        return str(self.to_dict())
    
    def __hash__(self):
        return hash((self.min, self.max, self.probable))
    
    def __eq__(self, other):
        if isinstance(other, MonteCarloRange) and self.__hash__() == other.__hash__():
            return True
        return False
    
    def __gt__(self, other):
        if self == other:
            return False
        if self.min >= other.min and self.max > other.max:
            return True
        return False

class PertDistribution():
    def __init__(self,range: MonteCarloRange):
        rng = np.random.default_rng()
        delta_min_max = range.max - range.min
        alpha = 1 + ((range.probable - range.min) * 4) / delta_min_max
        beta = 1 + ((range.max - range.probable) * 4) / delta_min_max
        self.__samples = float(range.min) + rng.beta(alpha, beta, 100000) * float(delta_min_max)

    def get(self):
        return self.__samples.copy()


class MonteCarloSimulation():
    def __init__(self, range: MonteCarloRange=None):
        if not range:
            range = MonteCarloRange(probable=Decimal(1))
        if (range.min == range.max == range.probable):
            self.probable = Decimal(range.probable)
            range.max = Decimal(Decimal(range.probable)+Decimal(0.000000000001))
            range.min = Decimal(Decimal(range.probable)-Decimal(0.000000000001))
        
        pd = PertDistribution(range=range)
        self.__samples = pd.get()
        self.probable = Decimal(np.mean(self.__samples))
        self.p90 = Decimal(np.percentile(self.__samples, 90))
        self.max = Decimal(np.max(self.__samples))
        self.min = Decimal(np.min(self.__samples))

    def to_dict(self):
        return {
            "min": float(self.min),
            "probable": float(self.probable),
            "max": float(self.max),
            "p90": float(self.p90),
            "__samples": self.__samples
        }
    
    @classmethod
    def from_dict(cls, values:dict={}):
        if isinstance(values, dict) and 'p90' in values:
            mcs = MonteCarloSimulation()
            mcs.min = Decimal(values['min'])
            mcs.probable = Decimal(values['probable'])
            mcs.max = Decimal(values['max'])
            mcs.p90 = Decimal(values['p90'])
            mcs.__samples = np.array(values.get('__samples', []))
            return mcs
        else:
            range = MonteCarloRange.from_dict(values)
            return MonteCarloSimulation(range=range)
    
    def get_montecarlo_range(self):
        return MonteCarloRange(min=self.min, probable=self.probable, max=self.max)
        
    def multiply(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().multiply(other=other.get_montecarlo_range())
        else:
            range = self.get_montecarlo_range().multiply(other=other)
        return MonteCarloSimulation(range=range)
    
    def add(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().add(other=other.get_montecarlo_range())
        else:
            range = self.get_montecarlo_range().add(other=other)
        return MonteCarloSimulation(range=range)
    
    def delta(self, other):
        if isinstance(other, MonteCarloSimulation):
            range = self.get_montecarlo_range().sub(other=other.get_montecarlo_range())
        else:
            range = self.get_montecarlo_range().sub(other=other)
        return MonteCarloSimulation(range=range)
    
    def __add__(self, other):
        return self.add(other)
    def __radd__(self, other):
        return self.add(other)
    def __mul__(self, other):
        return self.multiply(other)
    def __rmul__(self, other):
        return self.multiply(other)
    def __repr__(self):
        return str(self.to_dict())
    
    def __str__(self):
        return f"min: {str(self.min)} mode: {str(self.probable)} p90: {str(self.p90)} max: {str(self.max)}"
    
    def __hash__(self):
        return hash((self.min, self.probable, self.p90, self.max))
    
    def __eq__(self, value):
        if isinstance(value, MonteCarloSimulation) and self.__hash__() == value.__hash__():
            return True
        return False
    
    def __gt__(self, other):
        if self == other:
            return False
        if self.min >= other.min and self.max > other.max:
            return True
        return False


import unittest

from src.otyg_risk_base.qualitative_risk import QualitativeRisk



class TestQuantitativeRisk(unittest.TestCase):
    def test_output(self):
        risk = QualitativeRisk()
        expected = {'risk': "Very Low", 'likelihood': "Very Low", 'impact': "Very Low"}
        self.assertEqual(risk.get(), expected)
        risk = QualitativeRisk(likelihood_init=3, likelihood_impact=3, impact=5)
        expected = {'risk': "High", 'likelihood': "Moderate", 'impact': "Very High"}
        self.assertEqual(risk.get(), expected)

    def test_eq(self):
        risk = QualitativeRisk()
        riskII = QualitativeRisk(likelihood_init=3, likelihood_impact=3, impact=5)
        self.assertTrue(risk == risk)
        self.assertFalse(risk == riskII)
    
    def test_persistence(self):
        risk = QualitativeRisk()
        self.assertEqual(risk, QualitativeRisk.from_dict(risk.to_dict()))
        riskII = QualitativeRisk(likelihood_init=3, likelihood_impact=3, impact=5)
        self.assertEqual(riskII, QualitativeRisk.from_dict(riskII.to_dict()))

if __name__ == '__main__':
    unittest.main()


#
# BSD 3-Clause License
#
# Copyright (c) 2026, Martin Vesterlund
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from decimal import *

from .montecarlo import MonteCarloRange, MonteCarloSimulation


class QuantitativeRisk():
    def __init__(self, values: dict = None):
        if not values:
            self.threat_event_frequency = MonteCarloRange()
            self.vuln_score = MonteCarloRange()
            self.loss_magnitude = MonteCarloSimulation()
            self.budget = Decimal(1000000)
            self.currency = "SEK"
        else:
            self.threat_event_frequency = MonteCarloRange.from_dict(values['threat_event_frequency'])
            self.vuln_score = MonteCarloRange.from_dict(values['vulnerability'])
            self.loss_magnitude = MonteCarloSimulation.from_dict(values.get('loss_magnitude'))
            self.budget = Decimal(values['budget'])
            self.currency = values['currency']
        if values and 'loss_event_frequency' in values:
            self.loss_event_frequency = MonteCarloSimulation.from_dict(values.get('loss_event_frequency'))
            if 'ale' in values and 'annual_loss_expectancy' in values:
                self.ale = MonteCarloSimulation.from_dict(values.get('ale'))
                self.annual_loss_expectancy = MonteCarloSimulation.from_dict(values.get('annual_loss_expectancy'))
            else:
                self.update_ale()
        else:
            self.loss_event_frequency = MonteCarloSimulation(self.threat_event_frequency.multiply(self.vuln_score))
            self.update_ale()
    
    def update_ale(self):
        self.ale = self.loss_event_frequency.multiply(self.loss_magnitude).multiply(self.budget)
        self.annual_loss_expectancy = MonteCarloSimulation(self.ale)

    def to_dict(self):
        return {
            "threat_event_frequency": self.threat_event_frequency.to_dict(),
            "vulnerability": self.vuln_score.to_dict(),
            "loss_event_frequency": self.loss_event_frequency.to_dict(),
            "loss_magnitude": self.loss_magnitude.to_dict(),
            "ale": self.ale.to_dict(),
            "annual_loss_expectancy": self.annual_loss_expectancy.to_dict(),
            "budget": float(self.budget),
            "currency": self.currency
        }
    @classmethod
    def from_dict(cls, values):
        risk = QuantitativeRisk()
        risk.threat_event_frequency = MonteCarloRange.from_dict(values['threat_event_frequency'])
        risk.vuln_score = MonteCarloRange.from_dict(values['vulnerability'])
        risk.loss_magnitude = MonteCarloSimulation.from_dict(values.get('loss_magnitude'))
        risk.budget = Decimal(values['budget'])
        risk.currency = values['currency']
        risk.loss_event_frequency = MonteCarloSimulation.from_dict(values.get('loss_event_frequency'))
        risk.ale = MonteCarloSimulation.from_dict(values.get('ale'))
        risk.annual_loss_expectancy = MonteCarloSimulation.from_dict(values.get('annual_loss_expectancy'))
        return risk
    
    def __repr__(self):
        return str(self.to_dict())
    
    def __eq__(self, other):
        if isinstance(other, QuantitativeRisk):
            return self.__hash__() == other.__hash__()
        return False
    
    def __gt__(self, other):
        if self == other:
            return False
        res = False
        
    def __hash__(self):
        return hash((self.threat_event_frequency, self.vuln_score, self.loss_event_frequency, self.loss_magnitude, self.ale, self.annual_loss_expectancy, self.budget, self.currency))
    def __str__(self):
        return str("ALE (p90): " + str(round(self.annual_loss_expectancy.p90, 2)) + f" {self.currency}/år\nALE (p50): " + str(round(self.annual_loss_expectancy.probable, 2)) + f" {self.currency}/år\n")



def freeze(x):
    if isinstance(x, dict):
        return frozenset((k, freeze(v)) for k, v in x.items())
    if isinstance(x, (list, tuple)):
        return tuple(freeze(i) for i in x)
    if isinstance(x, set):
        return frozenset(freeze(i) for i in x)
    return x  # antar hashbart


from typing import Dict, List, Tuple

from .utils import freeze

class QualitativeScale():
    DEFAULT_NUM_TO_TEXT:Dict[str, int] = {
        5: 'Very High',
        4: 'High',
        3: 'Moderate',
        2: 'Low',
        1: 'Very Low',
        0: 'Very Low'
    }

    DEFAULT_RANGE_TO_TEXT: List[Dict] = [
        {'value':1, 'low': 1, 'high': 5},
        {'value':2, 'low': 5, 'high': 9},
        {'value':3, 'low': 9, 'high': 13},
        {'value':4, 'low': 13, 'high': 20},
        {'value':5, 'low': 20, 'high': 26},
    ]

    DEFAULT_TEF: List[Dict] = [
        {'value':1, 'low': 0.01, 'high': 0.1},
        {'value':2, 'low': 0.1, 'high': 1},
        {'value':3, 'low': 1, 'high': 10},
        {'value':4, 'low': 10, 'high': 100},
        {'value':5, 'low': 100, 'high': 1000},
    ]

    DEFAULT_VULNERABILITY: List[Dict] = [
        {'value':1, 'low': 0.01, 'high': 0.12},
        {'value':2, 'low': 0.12, 'high': 0.25},
        {'value':3, 'low': 0.25, 'high': 0.5},
        {'value':4, 'low': 0.5, 'high': 0.75},
        {'value':5, 'low': 0.75, 'high': 1},
    ]

    DEFAULT_CONSEQUENCE: List[Dict] = [
        {'value':1, 'low': 0.0001, 'high': 0.005},
        {'value':2, 'low': 0.005, 'high': 0.01},
        {'value':3, 'low': 0.01, 'high': 0.02},
        {'value':4, 'low': 0.02, 'high': 0.05},
        {'value':5, 'low': 0.05, 'high': 1},
    ]
    def __init__(self, scales:dict = None):
        if not scales:
            scales = dict()
        self.likelihood_initiation_or_occurence = scales.get('likelihood_initiation_or_occurence', self.DEFAULT_TEF)
        self.likelihood_adverse_impact = scales.get('likelihood_adverse_impact', self.DEFAULT_VULNERABILITY)
        self.impact = scales.get('impact', self.DEFAULT_CONSEQUENCE)
        self.num_to_text = scales.get('num_to_text', self.DEFAULT_NUM_TO_TEXT)
        self.risk = scales.get('risk', self.DEFAULT_RANGE_TO_TEXT)

    def to_dict(self):
        return {
            'likelihood_initiation_or_occurence': self.likelihood_initiation_or_occurence,
            'likelihood_adverse_impact': self.likelihood_adverse_impact,
            'impact': self.impact,
            'num_to_text': self.num_to_text,
            'risk': self.risk,
        }
    def __hash__(self):
        return hash(freeze(self.to_dict()))
    
    def __eq__(self, other):
        return self.__hash__() == other.__hash__()
    
    @classmethod
    def from_dict(cls, scales: dict):
        tmp = QualitativeScale()
        tmp.likelihood_initiation_or_occurence = scales.get('likelihood_initiation_or_occurence', cls.DEFAULT_TEF)
        tmp.likelihood_adverse_impact = scales.get('likelihood_adverse_impact', cls.DEFAULT_VULNERABILITY)
        tmp.impact = scales.get('impact', cls.DEFAULT_CONSEQUENCE)
        tmp.num_to_text = scales.get('num_to_text', cls.DEFAULT_NUM_TO_TEXT)
        tmp.risk = scales.get('risk', cls.DEFAULT_RANGE_TO_TEXT)
        return tmp
    
    def get(self, raw, mapping):
        scale = getattr(self, mapping, None)
        num = None
        if not scale:
            raise AttributeError(f"{mapping} does not exist")
        if raw < scale[0].get('low'):
            num = scale[0].get('value')
        elif raw > scale[-1].get('low'):
            num = scale[-1].get('value')
        else:
            for range in scale:
                if range.get('low') <= raw < range.get('high'):
                    num = range.get('value')
        lookup = num
        if not self.num_to_text.get(num):
            lookup = str(num)
        return {'numeric': num, 'text': self.num_to_text.get(lookup)}
